import numpy as np


def get_mean_cellscore_per_group(cell_scores, cell_groups):
    unique_cluster_ids = np.unique(cell_groups)
    mean_cluster_scores = []
    for c in unique_cluster_ids:
        cell_idx = np.where(cell_groups == c)[0]
        mean_cluster_scores.append(np.mean(cell_scores[cell_idx], axis=0))
    mean_cluster_scores = np.array(mean_cluster_scores)
    return mean_cluster_scores


def mod_score(factors_by_groups_matrix):
    total_factor_relative_weight = np.sum(factors_by_groups_matrix, axis=1)
    total_factor_relative_weight = total_factor_relative_weight / np.sum(
        total_factor_relative_weight
    )

    # Per factor
    n1 = (
        factors_by_groups_matrix
        / np.sum(factors_by_groups_matrix, axis=1)[:, np.newaxis]
    )
    n1 = np.sum(np.max(n1, axis=1) * total_factor_relative_weight)

    # Per group
    n2 = (
        factors_by_groups_matrix
        / np.sum(factors_by_groups_matrix, axis=0)[np.newaxis, :]
    )
    n2 = np.mean(np.max(n2, axis=0))

    return np.mean([n1, n2])


def entropy_score(factors_by_groups_matrix):
    total_factor_relative_weight = np.sum(factors_by_groups_matrix, axis=1)
    total_factor_relative_weight = total_factor_relative_weight / np.sum(
        total_factor_relative_weight
    )

    # Per factor
    n1 = (
        factors_by_groups_matrix
        / np.sum(factors_by_groups_matrix, axis=1)[:, np.newaxis]
    )
    n1 = np.sum(-np.sum(n1 * np.log(n1), axis=1) * total_factor_relative_weight)

    # Per group
    n2 = (
        factors_by_groups_matrix
        / np.sum(factors_by_groups_matrix, axis=0)[np.newaxis, :]
    )
    n2 = np.mean(-np.sum(n2 * np.log(n2), axis=0))

    return np.mean([n1, n2])


def compute_geneset_coherence(genes, counts_adata):
    # As in Spectra: https://github.com/dpeerlab/spectra/blob/ff0e5c456127a33938b1ea560432f228dc26a08b/spectra/initialization.py
    mat = np.array(counts_adata[:, genes].X)
    n_genes = len(genes)
    score = 0
    for i in range(1, n_genes):
        for j in range(i):
            dw1 = mat[:, i] > 0
            dw2 = mat[:, j] > 0
            dw1w2 = (dw1 & dw2).astype(float).sum()
            dw1 = dw1.astype(float).sum()
            dw2 = dw2.astype(float).sum()
            score += np.log((dw1w2 + 1) / (dw2))

    denom = n_genes * (n_genes - 1) / 2

    return score / denom


def coherence_score(marker_gene_sets, heldout_counts_adata):
    chs = []
    for marker_genes in marker_gene_sets:
        chs.append(compute_geneset_coherence(marker_genes, heldout_counts_adata))
    return np.mean(chs)


def jaccard_similarity(lists):
    list_of_sets = [set(list) for list in lists]
    intersection = set.intersection(*list_of_sets)
    union = set.union(*list_of_sets)
    return float(len(intersection) / float(len(union)))


def overlap_index(list1, list2):
    s1 = set(list1)
    s2 = set(list2)
    size1 = len(s1)
    size2 = len(s2)
    return float(len(s1.intersection(s2)) / float(min(size1, size2)))


def compute_fscore(tp, fp, fn):
    return 2 * tp / (2 * tp + fp + fn)


def score_signature(inf_signature, markers_list, nonmarkers_list):
    # true positive: marker for this type is in signature
    # true negative: marker for another type is not in signature
    tp = sum([1 for gene in inf_signature if gene in markers_list])
    fp = sum([1 for gene in inf_signature if gene in nonmarkers_list])
    fn = sum([1 for gene in markers_list if gene not in inf_signature])
    return compute_fscore(tp, fp, fn)


def compute_cluster_obs_association_score(
    adata, cluster_assignments, cluster_name, obs_key, obs_val
):
    # Cells attached to factor
    cells_in_factor = np.where(np.array(cluster_assignments) == cluster_name)[0]
    adata_cells_in_factor = adata[cells_in_factor]

    # Cells from obs_val
    adata_cells_from_obs = adata[np.where(adata.obs[obs_key] == obs_val)[0]]

    cells_from_obs = float(adata_cells_from_obs.shape[0])

    # Number of cells from obs_val that are not in factor
    cells_not_in_factor_from_obs = float(
        np.count_nonzero(
            list(
                set(adata_cells_from_obs.obs.index).difference(
                    set(adata_cells_in_factor.obs.index)
                )
            )
        )
    )

    # Number of cells in factor that are obs_val
    cells_in_factor_from_obs = float(
        np.count_nonzero(adata_cells_in_factor.obs[obs_key] == obs_val)
    )

    # Number of cells in factor that are not obs_val
    cells_in_factor_not_from_obs = float(
        np.count_nonzero(adata_cells_in_factor.obs[obs_key] != obs_val)
    )

    return compute_fscore(
        cells_in_factor_from_obs,
        cells_in_factor_not_from_obs,
        cells_not_in_factor_from_obs,
    )


def gini(x):
    # (Warning: This is a concise implementation, but it is O(n**2)
    # in time and memory, where n = len(x).  *Don't* pass in huge
    # samples!)

    # Mean absolute difference
    mad = np.abs(np.subtract.outer(x, x)).mean()
    # Relative mean absolute difference
    rmad = mad / np.mean(x)
    # Gini coefficient
    g = 0.5 * rmad
    return g



import numpy as np
import matplotlib.pyplot as plt

def plot_cell_entropies(self, thres=0.9):
    entropies = []
    factor_ns = []
    for layer_idx in range(4):
        layer_name = self.layer_names[layer_idx]
        nf = len(self.factor_lists[layer_idx])
        a = self.adata.obsm[f"X_{layer_name}factors"] / np.sum(self.adata.obsm[f"X_{layer_name}factors"],axis=1)[:,None]
        a_sorted = np.vstack([a[i,np.argsort(a,axis=1)[:,::-1][i]] for i in range(a.shape[0])])
        a_cumsums = np.cumsum(a_sorted,axis=1)
        n_factors = np.array([(np.where(a_cumsums[i] > thres)[0][0] + 1) for i in range(a.shape[0])])
        factor_ns.append(n_factors)
        entropy = np.array([np.sum(-np.log(a[i])*a[i])/np.log(nf) for i in range(a.shape[0])])
        entropies.append(entropy)
    fig, axes = plt.subplots(1,2)
    plt.sca(axes[0])
    plt.boxplot(entropies)
    plt.sca(axes[1])
    plt.boxplot(factor_ns)
    plt.show()

def plot_factor_genes(self, thres=0.9):
    layer_kept_n_genes = []
    layer_removed_n_genes = []
    factor_n_kept_genes = dict()
    for layer_idx in range(len(self.layer_sizes)):
        layer_name = f'{self.layer_names[layer_idx]}'
        
        term_scores = self.pmeans[f"{self.layer_names[0]}W"]
        
        if layer_idx > 0:
            term_scores = self.pmeans[f"{self.layer_names[layer_idx]}W"]
            for layer in range(layer_idx - 1, 0, -1):
                lower_mat = self.pmeans[f"{self.layer_names[layer]}W"]
                term_scores = term_scores.dot(lower_mat)
            term_scores = term_scores.dot(
                self.pmeans[f"{self.layer_names[0]}W"]
            )
        
        kept_factors_n_genes = []
        removed_factors_n_genes = []
        f_idx = 0
        for factor in range(self.layer_sizes[layer_idx]):
            vals = term_scores[factor] / np.sum(term_scores[factor])
            # sort
            vals_sorted = vals[np.argsort(vals)[::-1]]
            if factor in self.factor_lists[layer_idx]:
                kept_factors_n_genes.append(np.where(np.cumsum(vals_sorted) > thres)[0][0])
                factor_n_kept_genes[f'{layer_name}{f_idx}'] = kept_factors_n_genes[-1]
                f_idx += 1
            else:
                removed_factors_n_genes.append(np.where(np.cumsum(vals_sorted) > thres)[0][0])
            
        layer_kept_n_genes.append(kept_factors_n_genes)
        layer_removed_n_genes.append(removed_factors_n_genes)

    m = np.max(list(factor_n_kept_genes.values()))
    for f in factor_n_kept_genes:
        factor_n_kept_genes[f] = factor_n_kept_genes[f] / m
    factor_n_kept_genes

    fig, axes = plt.subplots(1,4, figsize=(6,3), sharey=True)
    for i in range(4):
        plt.sca(axes[i])
        plt.boxplot([layer_kept_n_genes[i], layer_removed_n_genes[i]], tick_labels=['Kept', 'Removed'])
    plt.suptitle('Number of genes in factors')

def plot_factor_gini(self, idx, thres=0.9):
    a = self.pmeans[f"{self.layer_names[0]}W"][idx]
    # a -= np.mean(a)
    # a = np.maximum(a, 0)
    norm_a = a/np.sum(a)
    vals_sorted = norm_a[np.argsort(norm_a)[::-1]]
    plt.plot(np.cumsum(vals_sorted))
    plt.axvline(np.where(np.cumsum(vals_sorted) > thres)[0][0], color='gray')
    # plt.plot(a)
    plt.title(self.pmeans['brd'][idx])
    plt.show()    


import jax
import jax.numpy as jnp
from jax import vmap
import tensorflow_probability.substrates.jax.distributions as tfd


def gamma_sample(rng, shape, rate):
    scale = 1.0 / rate
    return jnp.clip(
        tfd.Gamma(shape, rate).sample(seed=rng), a_min=1e-15, a_max=1e15
    )


def _gamma_logpdf(x, shape, rate):
    return tfd.Gamma(shape, rate).log_prob(x)


def gamma_logpdf(x, shape, rate):
    shape = shape * jnp.ones(x.shape)
    rate = rate * jnp.ones(x.shape)
    return jnp.sum(
        vmap(vmap(_gamma_logpdf, in_axes=(0, 0, 0)), in_axes=(0, 0, 0))(x, shape, rate)
    )


def _gamma_entropy(shape, rate):
    return tfd.Gamma(shape, rate).entropy()


def gamma_entropy(shape, rate):
    return jnp.sum(
        vmap(vmap(_gamma_entropy, in_axes=(0, 0)), in_axes=(0, 0))(shape, rate)
    )

def lognormal_sample(rng, m, s):
    return jnp.clip(tfd.LogNormal(m, s).sample(seed=rng), 1e-15, 1e15)

def _lognormal_entropy(m, s):
    return tfd.LogNormal(m, s).entropy()

def lognormal_entropy(m, s):
    return jnp.sum(
        vmap(vmap(_lognormal_entropy, in_axes=(0, 0)), in_axes=(0, 0))(m, s)
    )

def _lognormal_logpdf(x, m, s):
    return tfd.LogNormal(m, s).log_prob(x)

def lognormal_logpdf(x, m, s):
    m = m * jnp.ones(x.shape)
    s = s * jnp.ones(x.shape)
    return jnp.sum(
        vmap(vmap(_lognormal_logpdf, in_axes=(0, 0, 0)), in_axes=(0, 0, 0))(x, m, s)
    )


from scdef.utils.score_utils import *

from collections import ChainMap


def get_hierarchy_from_clusters(clusters_levels, level_prefix="h", use_names=False):
    hierarchy = dict()
    level_prefixes = [level_prefix * level for level in range(len(clusters_levels))]
    for level, clusters in enumerate(clusters_levels[:-1]):
        clusters_up = clusters_levels[level + 1]
        unique_clusters = np.unique(clusters)
        unique_clusters_up = np.unique(clusters_up)
        for i, cluster in enumerate(unique_clusters):
            if use_names:
                name = f"{cluster}"
            else:
                name = f"{level_prefixes[level]}{i}"

            # Get all cells attached to this cluster
            cells_a = np.where(np.array(clusters) == cluster)[0]
            scores = []
            for j, cluster_up in enumerate(unique_clusters_up):
                cells_b = np.where(np.array(clusters_up) == cluster_up)[0]
                scores.append(jaccard_similarity([cells_a, cells_b]))

            if use_names:
                upper_level = f"{unique_clusters_up[np.argmax(scores)]}"
            else:
                upper_level = f"{level_prefixes[level+1]}{np.argmax(scores)}"

            if upper_level in hierarchy:
                hierarchy[upper_level].append(name)
            else:
                hierarchy[upper_level] = []
                hierarchy[upper_level].append(name)

    return hierarchy


def flatten_hierarchy(hierarchy_dict):
    def descend(factor):
        node = [factor]
        if factor in hierarchy_dict:
            children = list(hierarchy_dict[factor])
            for child in children:
                nodes = descend(child)
                node.extend(nodes)
        return node

    flattened = {}
    for factor in hierarchy_dict:
        flattened[factor] = descend(factor)[1:]

    return flattened


def get_nodes_from_hierarchy(hierarchy):
    hierarchy_nodes = None
    hierarchy_nodes = list(hierarchy.keys())
    vals = [item for sublist in list(hierarchy.values()) for item in sublist]
    hierarchy_nodes.extend(vals)
    hierarchy_nodes = list(set(hierarchy_nodes))  # remove duplicates
    return hierarchy_nodes


# Annotate hierarchy
def annotate_hierarchy(hierarchy, factor_annotation_matches):
    annotated_hierarchy = {}
    for hrc in hierarchy:
        factors = hierarchy[hrc]
        factor_annots = []
        for factor in factors:
            if factor in factor_annotation_matches:
                factor_annots.append(factor_annotation_matches[factor])
            else:
                factor_annots.append(factor)
        if hrc in factor_annotation_matches:
            annotated_hierarchy[factor_annotation_matches[hrc]] = factor_annots
        else:
            annotated_hierarchy[hrc] = factor_annots
    return annotated_hierarchy


# Compare hierarchies
def compare_hierarchies(inf_complete, true_complete):
    # Need the hierarchies to have all types
    flattened_inferred = flatten_hierarchy(inf_complete)
    flattened_true = flatten_hierarchy(true_complete)
    levels = list(flattened_true.keys())

    scores = []
    for hrc in levels:
        true_below = flattened_true[hrc]
        true_not_below = []
        for h in levels:
            if h != hrc and h not in true_below:
                true_not_below.append(h)
                for down in flattened_true[h]:
                    true_not_below.append(down)
        true_not_below = list(set(true_not_below))

        # Get inferred set of types below hrc
        inf_below = flattened_inferred[hrc]

        # true positive: marker for this type is in signature
        # true negative: marker for another type is not in signature
        tp = sum([1 for grp in inf_below if grp in true_below])
        fp = sum([1 for grp in inf_below if grp in true_not_below])
        fn = sum([1 for grp in true_below if grp not in inf_below])
        if fp == 0 and len(true_below) == 0:
            # the inf hierarchy does not disagree with the true but it also doesn't have TP because the true hierarchy is already flat
            score = 1
        else:
            score = compute_fscore(tp, fp, fn)

        scores.append(score)

    return np.mean(scores)


def simplify_hierarchy(hierarchy, layer_names, layer_sizes, factor_names=None):
    simplified = hierarchy.copy()
    n_layers = len(layer_names)
    for layer_idx in range(1, n_layers):
        layer_name = layer_names[layer_idx]
        for factor_idx in range(layer_sizes[layer_idx]):
            if factor_names is not None:
                factor_name = factor_names[layer_idx][factor_idx]
            else:
                factor_name = f"{layer_name}{factor_idx}"
            if factor_name in simplified:
                if len(simplified[factor_name]) == 1:
                    if simplified[factor_name][0] in simplified:
                        down_hrc = simplified[simplified[factor_name][0]]
                        old_hrc = simplified[factor_name][0]
                        if len(down_hrc) > 1:
                            simplified[factor_name] = down_hrc
                            del simplified[old_hrc]
                        else:
                            if layer_idx < n_layers - 1:
                                del simplified[factor_name]
                            else:
                                simplified[factor_name] = []
                    else:
                        del simplified[factor_name]

    return simplified


def get_cluster_obs_association_scores(adata, clusters, layer_names, obs_key, obs_val):
    scores = []
    factors = []
    layers = []
    n_layers = len(clusters)
    for layer_idx in range(n_layers):
        layer_name = layer_names[layer_idx]
        n_clusters = len(np.unique(clusters[layer_idx]))
        cluster_assignments = clusters[layer_idx]
        for factor_idx in range(n_clusters):
            cluster_name = f"{layer_name}{factor_idx}"
            score = compute_cluster_obs_association_score(
                adata, cluster_assignments, cluster_name, obs_key, obs_val
            )
            scores.append(score)
            factors.append(cluster_name)
            layers.append(layer_idx)
    return scores, factors, layers


def assign_obs_to_clusters(
    adata, clusters_levels, layer_names, obs_keys, cluster_names=[]
):
    if not isinstance(obs_keys, list):
        obs_keys = [obs_keys]

    obs_to_factor_assignments = []
    obs_to_factor_matches = []
    for obs_key in obs_keys:
        obskey_to_factor_assignments = dict()
        obskey_to_factor_matches = dict()
        for obs in adata.obs[obs_key].unique():
            scores, factors, layers = get_cluster_obs_association_scores(
                adata, clusters_levels, layer_names, obs_key, obs
            )
            if len(cluster_names) > 0:
                # Subset to factor_names
                idx = np.array(
                    [i for i, factor in enumerate(factors) if factor in cluster_names]
                )
                scores = np.array(scores)[idx]
                factors = np.array(factors)[idx]
                layers = np.array(layers)[idx]
            obskey_to_factor_assignments[obs] = factors[np.argmax(scores)]
            obskey_to_factor_matches[factors[np.argmax(scores)]] = obs
        obs_to_factor_assignments.append(obskey_to_factor_assignments)
        obs_to_factor_matches.append(obskey_to_factor_matches)

    # Join them all up
    factor_annotation_assignments = ChainMap(*obs_to_factor_assignments)
    factor_annotation_matches = ChainMap(*obs_to_factor_matches)

    return dict(factor_annotation_assignments), dict(factor_annotation_matches)


def complete_hierarchy(hierarchy, obs_vals):
    new_hierarchy = hierarchy.copy()
    for obs_val in obs_vals:
        if obs_val not in hierarchy:
            new_hierarchy[obs_val] = []
    return new_hierarchy



from scdef.utils import score_utils, hierarchy_utils, color_utils
from scdef.utils.jax_utils import *

from jax import jit, grad, vmap
from jax.example_libraries import optimizers
from jax import random, value_and_grad
from jax.scipy.stats import norm, gamma, poisson
import jax.numpy as jnp
import jax.nn as jnn
import jax

from scdef.models._scdef import scDEF

from anndata import AnnData
import jax.numpy as jnp
import numpy as np
import logging

from typing import Optional, Sequence, Mapping


class iscDEF(scDEF):
    """Informed scDEF model.

    This model extends the basic scDEF by using gene sets to guide the factors.
    iscDEF can either set the given sets as top layer factors and learn higher-resolution
    structure, or use them as the lowest resolution and learn a hierarchy that relates them.
    All the methods from scDEF are available in iscDEF.

    Args:
        adata: AnnData object containing the gene expression data. scDEF learns a model from
            counts, so they must be present in either adata.X or in adata.layers.
        markers_dict: dictionary containing named gene lists.
        add_other: whether to add factors for cells which don't express any of the sets in markers_dict.
        markers_layer: scDEF layer at which the gene sets are defined. If > 0, this defines the number of layers.
        n_factors_per_set: number of lower level factors per gene set.
        n_layers: default number of scDEF layers, including a top layer of size 1 if markers_layer is 0.
        cn_small_scale: scale for low connectivity
        cn_big_scale: scale for large connectivity
        cn_small_strength: strength for weak connectivity
        cn_big_strength: strength for large connectivity
        gs_small_scale: scale for genes not in set
        gs_big_scale: scale for genes in set
        marker_strength: strength for marker genes
        nonmarker_strength: strength for non-marker genes
        other_strength: strength for marker genes of other sets
        **kwargs: keyword arguments for base scDEF.
    """

    def __init__(
        self,
        adata: AnnData,
        markers_dict: Mapping[str, Sequence[str]],
        add_other: Optional[
            bool
        ] = False,  # whether to add factors for cells which don't express any of the sets in markers_matrix
        markers_layer: Optional[
            int
        ] = 0,  # by default, use lower layer and learn a hierarchy
        n_factors_per_set: Optional[int] = 3,
        n_sets_per_factor: Optional[int] = 1.5,
        n_layers: Optional[int] = 4,
        cn_small_mean: Optional[float] = 0.01,
        cn_big_mean: Optional[float] = 10.0,
        cn_small_strength: Optional[float] = 100.0,
        cn_big_strength: Optional[float] = 1.0,
        gs_small_scale: Optional[float] = 0.1,
        gs_big_scale: Optional[float] = 10.0,
        marker_strength: Optional[float] = 100.0,
        nonmarker_strength: Optional[float] = 1.0,
        other_strength: Optional[float] = 100.0,
        **kwargs,
    ):
        self.markers_dict = markers_dict
        self.add_other = add_other
        self.markers_layer = markers_layer
        self.n_layers = n_layers
        self.n_factors_per_set = n_factors_per_set
        self.n_sets_per_factor = n_sets_per_factor
        self.gs_big_scale = gs_big_scale
        self.cn_big_strength = cn_big_strength

        self.marker_names = list(self.markers_dict.keys()) + ["other"] * self.add_other
        self.n_markers = len(self.marker_names)

        layer_sizes = []
        layer_names = []
        if markers_layer == 0:
            for layer in range(self.n_layers):
                if layer == 0:
                    size = self.n_markers
                    name = "marker"
                else:
                    size = int(
                        np.ceil(self.n_markers / (self.n_sets_per_factor * layer))
                    )
                    name = "h" * layer
                    if size <= 1:
                        break
                layer_sizes.append(size)
                layer_names.append(name)
            layer_sizes.append(1)
            layer_names.append("h" * (layer + 1))
            self.n_layers = len(layer_sizes)

            if "layer_shapes" not in kwargs:
                kwargs["layer_shapes"] = [0.3] * (self.n_layers - 1) + [1.0]
            if "factor_shapes" not in kwargs:
                kwargs["factor_shapes"] = [1.0] * self.n_layers
            if "factor_rates" not in kwargs:
                kwargs["factor_rates"] = [1.0] * self.n_layers

            if "use_brd" not in kwargs:
                kwargs["use_brd"] = False
            elif kwargs["use_brd"] == True:
                raise ValueError("`use_brd` must be False if markers_layer is 0")
        else:
            self.n_layers = self.markers_layer + 1
            for layer in range(self.n_layers):
                if layer < self.n_layers - 1:
                    rev_layer = (self.n_layers - 1) - layer
                    size = self.n_markers * self.n_factors_per_set * rev_layer
                    name = "h" * layer
                else:
                    size = self.n_markers
                    name = "marker"
                layer_sizes.append(size)
                layer_names.append(name)

            if "layer_shapes" not in kwargs:
                kwargs["layer_shapes"] = 0.3
            if "factor_shapes" not in kwargs:
                kwargs["factor_shapes"] = 1.0
            if "factor_rates" not in kwargs:
                kwargs["factor_rates"] = 1.0

        super(iscDEF, self).__init__(adata, layer_sizes=layer_sizes, **kwargs)

        logginglevel = self.logger.level
        self.logger = logging.getLogger("iscDEF")
        self.logger.setLevel(logginglevel)

        self.layer_names = layer_names

        # Set w_priors
        if self.markers_layer != 0:
            self.set_connectivity_prior(
                cn_small_strength=cn_small_strength,
                cn_big_strength=cn_big_strength,
                cn_small_mean=cn_small_mean,
                cn_big_mean=cn_big_mean,
            )

        # Do gene sets
        self.set_geneset_prior(
            gs_big_scale=gs_big_scale,
            gs_small_scale=gs_small_scale,
            marker_strength=marker_strength,
            nonmarker_strength=nonmarker_strength,
            other_strength=other_strength,
        )

        self.init_var_params()
        self.set_posterior_means()
        self.set_factor_names()

    def __repr__(self):
        out = f"iscDEF object with {self.n_layers} layers"
        out += "\n\t" + "Markers layer: " + str(self.markers_layer)
        out += "\n\t" + "Contains `other` category: " + str(self.add_other)
        out += "\n\t" + "Gene set strength: " + str(self.marker_strength)
        out += "\n\t" + "Gene set mean: " + str(self.gs_big_scale)
        out += (
            "\n\t"
            + "Layer names: "
            + ", ".join([f"{name}factor" for name in self.layer_names])
        )
        out += (
            "\n\t"
            + "Layer sizes: "
            + ", ".join([str(len(factors)) for factors in self.factor_lists])
        )
        out += (
            "\n\t"
            + "Layer shape parameters: "
            + ", ".join([str(shape) for shape in self.layer_shapes])
        )
        out += (
            "\n\t"
            + "Layer rate parameters: "
            + ", ".join([str(rate) for rate in self.layer_rates])
        )
        if self.markers_layer == 0:
            out += (
                "\n\t"
                + "Layer factor shape parameters: "
                + ", ".join([str(shape) for shape in self.factor_shapes])
            )
            out += (
                "\n\t"
                + "Layer factor rate parameters: "
                + ", ".join([str(rate) for rate in self.factor_rates])
            )
        else:
            out += "\n\t" + "Connectivity mean: " + str(self.cn_big_mean)
        out += "\n\t" + "Number of batches: " + str(self.n_batches)
        out += "\n" + "Contains " + self.adata.__str__()
        return out

    def set_connectivity_prior(
        self,
        cn_small_strength=100.0,
        cn_big_strength=1.0,
        cn_small_mean=0.01,
        cn_big_mean=1.0,
    ):
        self.cn_small_strength = cn_small_strength
        self.cn_big_strength = cn_big_strength
        self.cn_small_mean = cn_small_mean
        self.cn_big_mean = cn_big_mean

        # Do connectivities
        for layer_idx in range(1, self.n_layers):
            connectivity_matrix = cn_small_mean * np.ones(
                (self.layer_sizes[layer_idx], self.layer_sizes[layer_idx - 1])
            )
            strength_matrix = cn_small_strength * np.ones(
                (self.layer_sizes[layer_idx], self.layer_sizes[layer_idx - 1])
            )

            layer_rev_idx = self.n_layers - 1 - layer_idx

            if layer_idx == self.n_layers - 1:
                n_local_factors_per_set = 1
            else:
                n_local_factors_per_set = self.n_factors_per_set * (layer_rev_idx)
            n_lower_factors_per_set = self.n_factors_per_set * (layer_rev_idx + 1)
            for i in range(len(self.marker_names)):
                upper_start = i * n_local_factors_per_set
                upper_end = (i + 1) * n_local_factors_per_set

                local_start = i * n_lower_factors_per_set
                local_end = (i + 1) * n_lower_factors_per_set

                connectivity_matrix[
                    upper_start:upper_end, local_start:local_end
                ] = cn_big_mean

                strength_matrix[
                    upper_start:upper_end, local_start:local_end
                ] = cn_big_strength

            self.w_priors[layer_idx][0] = strength_matrix
            self.w_priors[layer_idx][1] = strength_matrix / connectivity_matrix

    def set_geneset_prior(
        self,
        gs_big_scale=10.0,
        gs_small_scale=0.1,
        marker_strength=100.0,
        nonmarker_strength=1.0,
        other_strength=100.0,
    ):
        self.gs_big_scale = gs_big_scale
        self.gs_small_scale = gs_small_scale
        self.marker_strength = marker_strength
        self.nonmarker_strength = nonmarker_strength
        self.other_strength = other_strength

        # Do gene sets
        self.gene_sets = np.ones((self.layer_sizes[0], self.n_genes)) * gs_small_scale
        self.strengths = (
            np.ones((self.layer_sizes[0], self.n_genes)) * nonmarker_strength
        )
        self.marker_gene_locs = []
        for i, cellgroup in enumerate(self.marker_names):
            factors_start = i
            factors_end = i + 1
            if self.markers_layer != 0:
                factors_start = i * self.n_factors_per_set * (self.n_layers - 1)
                factors_end = (i + 1) * self.n_factors_per_set * (self.n_layers - 1)

            if cellgroup != "other":
                for gene in self.markers_dict[cellgroup]:
                    loc = np.where(self.adata.var.index == gene)[0]
                    if len(loc) == 0:
                        self.logger.warning(
                            f"Did not find gene {gene} for set {cellgroup} in AnnData object."
                        )
                    self.marker_gene_locs.append(loc)
                    self.gene_sets[factors_start:factors_end, loc] = self.gs_big_scale
                    self.strengths[factors_start:factors_end, loc] = marker_strength

            # Make it hard for the factors in this group to give weight to genes in another group
            for group in self.markers_dict:
                if group != cellgroup:
                    for gene in self.markers_dict[group]:
                        if cellgroup != "other":
                            if gene not in self.markers_dict[cellgroup]:
                                loc = np.where(self.adata.var.index == gene)[0]
                                self.gene_sets[factors_start:factors_end, loc] = (
                                    gs_small_scale / 100.0
                                )
                                self.strengths[
                                    factors_start:factors_end, loc
                                ] = other_strength
                        else:
                            loc = np.where(self.adata.var.index == gene)[0]
                            self.gene_sets[factors_start:factors_end, loc] = (
                                gs_small_scale / 100.0
                            )
                            self.strengths[
                                factors_start:factors_end, loc
                            ] = other_strength

        self.w_priors[0][0] = jnp.array(self.strengths)  # shape
        self.w_priors[0][1] = jnp.array(self.strengths / self.gene_sets)  # rate

    def set_factor_names(self):
        self.factor_names = []

        for idx in range(self.n_layers):
            layer_name = self.layer_names[idx]
            if self.markers_layer == 0:
                if idx == 0:
                    self.factor_names.append(
                        [f"{self.marker_names[i]}" for i in self.factor_lists[idx]]
                    )
                else:
                    self.factor_names.append(
                        [
                            f"{self.layer_names[idx]}{str(i)}"
                            for i in range(len(self.factor_lists[idx]))
                        ]
                    )
            else:  # if not zero, it's the top one, so append gene set names to each subsequent layer
                if idx == self.n_layers - 1:
                    factor_names = [
                        marker
                        for i, marker in enumerate(self.marker_names)
                        if i in self.factor_lists[idx]
                    ]
                else:
                    rev_idx = self.n_layers - 1 - idx
                    factor_names = []
                    for marker_idx, marker_name in enumerate(self.marker_names):
                        marker_factor_names = []
                        sub_factors = np.arange(
                            marker_idx * self.n_factors_per_set * rev_idx,
                            (marker_idx + 1) * self.n_factors_per_set * rev_idx,
                        )
                        filtered_sub_factors = [
                            factor
                            for factor in sub_factors
                            if factor in self.factor_lists[idx]
                        ]
                        for sub_factor in range(len(filtered_sub_factors)):
                            marker_factor_names.append(
                                marker_name + f"_{self.layer_names[idx]}{sub_factor}"
                            )
                        factor_names += marker_factor_names
                self.factor_names.append(factor_names)

    def elbo(
        self,
        rng,
        batch,
        indices,
        var_params,
        annealing_parameter,
        stop_gradients,
        min_shape=1e-6,
        min_rate=1e-6,
        max_rate=1e6,
    ):
        # Single-sample Monte Carlo estimate of the variational lower bound.
        batch_indices_onehot = self.batch_indices_onehot[indices]

        cell_budget_params = var_params[0]
        gene_budget_params = var_params[1]
        fscale_params = var_params[2]
        z_params = var_params[3]

        cell_budget_shape = jnp.maximum(
            jnp.exp(cell_budget_params[0][indices]), min_shape
        )
        cell_budget_rate = jnp.minimum(
            jnp.maximum(jnp.exp(cell_budget_params[1][indices]), min_rate), max_rate
        )
        cell_budget_shape = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(cell_budget_shape),
            lambda: cell_budget_shape,
        )
        cell_budget_rate = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(cell_budget_rate),
            lambda: cell_budget_rate,
        )

        gene_budget_shape = jnp.maximum(jnp.exp(gene_budget_params[0]), min_shape)
        gene_budget_rate = jnp.minimum(
            jnp.maximum(jnp.exp(gene_budget_params[1]), min_rate), max_rate
        )
        gene_budget_shape = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(gene_budget_shape),
            lambda: gene_budget_shape,
        )
        gene_budget_rate = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(gene_budget_rate),
            lambda: gene_budget_rate,
        )

        fscale_shapes = jnp.maximum(jnp.exp(fscale_params[0]), min_shape)
        fscale_rates = jnp.minimum(
            jnp.maximum(jnp.exp(fscale_params[1]), min_rate), max_rate
        )
        fscale_shapes = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(fscale_shapes),
            lambda: fscale_shapes,
        )
        fscale_rates = jax.lax.cond(
            stop_gradients[0],
            lambda: jax.lax.stop_gradient(fscale_rates),
            lambda: fscale_rates,
        )

        z_shapes = jnp.maximum(jnp.exp(z_params[0][indices]), min_shape)
        z_rates = jnp.minimum(
            jnp.maximum(jnp.exp(z_params[1][indices]), min_rate), max_rate
        )

        # Sample from variational distribution
        cell_budget_sample = gamma_sample(rng, cell_budget_shape, cell_budget_rate)
        gene_budget_sample = gamma_sample(rng, gene_budget_shape, gene_budget_rate)
        if self.use_brd:
            fscale_samples = gamma_sample(rng, fscale_shapes, fscale_rates)
        # z_samples = gamma_sample(rng, z_shapes, z_rates)  # vectorized
        # w will be sampled in a loop below because it cannot be vectorized

        # Compute ELBO
        global_pl = gamma_logpdf(
            gene_budget_sample,
            self.gene_scale_shape,
            self.gene_scale_shape * self.gene_ratio,
        )
        global_en = -gamma_logpdf(
            gene_budget_sample, gene_budget_shape, gene_budget_rate
        )
        local_pl = gamma_logpdf(
            cell_budget_sample,
            self.cell_scale_shape,
            self.cell_scale_shape * self.batch_lib_ratio[indices],
        )
        local_en = -gamma_logpdf(
            cell_budget_sample, cell_budget_shape, cell_budget_rate
        )

        # scale
        if self.use_brd:
            global_pl += gamma_logpdf(fscale_samples, self.brd, self.brd * 100.0)
            global_en += -gamma_logpdf(fscale_samples, fscale_shapes, fscale_rates)

        z_mean = 1.0
        for idx in list(np.arange(0, self.n_layers)[::-1]):
            start = np.sum(self.layer_sizes[:idx]).astype(int)
            end = start + self.layer_sizes[idx]

            # w
            _w_shape = jnp.maximum(jnp.exp(var_params[4 + idx][0]), min_shape)
            _w_rate = jnp.minimum(
                jnp.maximum(jnp.exp(var_params[4 + idx][1]), min_rate), max_rate
            )

            _w_shape = jax.lax.cond(
                stop_gradients[idx],
                lambda: jax.lax.stop_gradient(_w_shape),
                lambda: _w_shape,
            )
            _w_rate = jax.lax.cond(
                stop_gradients[idx],
                lambda: jax.lax.stop_gradient(_w_rate),
                lambda: _w_rate,
            )

            _w_sample = gamma_sample(rng, _w_shape, _w_rate)
            if idx == 0 and self.use_brd:
                global_pl += gamma_logpdf(
                    _w_sample,
                    self.w_priors[idx][0] / fscale_samples,
                    self.w_priors[idx][1] / fscale_samples,
                )
            elif idx == 1 and self.use_brd:
                global_pl += gamma_logpdf(
                    _w_sample,
                    self.w_priors[idx][0],
                    self.w_priors[idx][1] / fscale_samples.T,
                )
            else:
                global_pl += gamma_logpdf(
                    _w_sample,
                    self.w_priors[idx][0],
                    self.w_priors[idx][1],
                )
            global_en += -gamma_logpdf(_w_sample, _w_shape, _w_rate)

            # z
            # _z_sample = z_samples[:, start:end]
            _z_shape = z_shapes[:, start:end]
            _z_rate = z_rates[:, start:end]

            _z_shape = jax.lax.cond(
                stop_gradients[idx],
                lambda: jax.lax.stop_gradient(_z_shape),
                lambda: _z_shape,
            )
            _z_rate = jax.lax.cond(
                stop_gradients[idx],
                lambda: jax.lax.stop_gradient(_z_rate),
                lambda: _z_rate,
            )
            _z_sample = gamma_sample(rng, _z_shape, _z_rate)

            if idx == self.n_layers - 1:
                local_pl += gamma_logpdf(
                    _z_sample, self.layer_shapes[idx], self.layer_rates[idx]
                )
            else:
                rate_param = self.layer_rates[idx]
                rate_param = jax.lax.cond(
                    stop_gradients[idx + 1],
                    lambda: rate_param * 0.0 + 1.0,
                    lambda: rate_param,
                )
                local_pl += gamma_logpdf(
                    _z_sample, self.layer_shapes[idx], rate_param / z_mean
                )
            local_en += -gamma_logpdf(_z_sample, _z_shape, _z_rate)

            z_mean = jnp.einsum("nk,kp->np", _z_sample, _w_sample)

            z_mean = jax.lax.cond(
                stop_gradients[idx], lambda: z_mean * 0.0 + 1, lambda: z_mean
            )

        # Compute log likelihood
        mean_bottom = jnp.einsum(
            "nk,kg->ng", _z_sample / cell_budget_sample, _w_sample
        ) / (batch_indices_onehot.dot(gene_budget_sample))
        ll = jnp.sum(vmap(poisson.logpmf)(jnp.array(batch), mean_bottom))

        # Anneal the entropy
        global_en *= annealing_parameter
        local_en *= annealing_parameter

        return ll + (
            local_pl
            + local_en
            + (indices.shape[0] / self.X.shape[0]) * (global_pl + global_en)
        )
